module.exports = function(Velocity, utils){
  var BLOCK_TYPES = ['if', 'foreach', 'macro', 'noescape', 'define'];
  /**
   * compile
   */
  utils.mixin(Velocity.prototype, {
    init: function(){
      this.context = {};
      this.macros = {};
      this.conditions = [];
      this.local = {};
      this.silence = false;

      utils.forEach(this.asts, this._init, this);
    },

    _init: function(ast, i){
      if (!ast.type || ast.type !== 'references') {
        this._trim(i + 1);
      }
    },

    /**
     * 删除多余的换行符，规则，所有非references的指令后面的换行符，都去除接下来的
     * 换行符
     */
    _trim: function(i){
      var asts = this.asts;
      var _ast = asts[i];
      if (typeof _ast === 'string' && _ast.slice(0, 1) === "\n") {
        asts[i] = _ast.slice(1);
      }
    },

    /**
     * @param context {object} 上下文环境，数据对象
     * @param macro   {object} self defined #macro
     * @param silent {bool} 如果是true，$foo变量将原样输出
     * @return str
     */
    render: function(context, macros, silence){

      this.silence = !!silence;
      this.context = context || {};
      this.jsmacros = macros || {};
      var t1 = utils.now();
      var str = this._render();
      var t2 = utils.now();
      var cost = t2 - t1;

      this.cost = cost;

      return str ;
    },

    /**
     * 解析入口函数
     * @param ast {array} 模板结构数组
     * @param contextId {number} 执行环境id，对于macro有局部作用域，变量的设置和
     * 取值，都放在一个this.local下，通过contextId查找
     * @return {string}解析后的字符串
     */
    _render: function(asts, contextId){

      var str = '';
      var block = [];
      var index = 0;
      asts = asts || this.asts;

      if (contextId) {

        if (contextId !== this.condition && 
            utils.indexOf(contextId, this.conditions) === -1) {
          this.conditions.unshift(contextId);
        }

        this.condition = contextId;

      } else {
        this.condition = null;
      }

      utils.forEach(asts, function(ast){
        var type = ast.type;

        //foreach if macro时，index加一
        if (utils.indexOf(type, BLOCK_TYPES) > -1) index ++;

        if (type === 'comment') return;

        if (index) {
          type === 'end' && index--;
          if (index) {
            block.push(ast);
            return;
          }
        }

        switch(type) {
          case 'references':
          str += this.getReferences(ast, true);
          break;

          case 'set':
          this.setValue(ast);
          break;

          case 'break':
          this.setBreak = true;
          break;

          case 'macro_call':
          str += this.getMacro(ast);
          break;

          case 'end':
          //使用slide获取block的拷贝
          str += this.getBlock(block.slice());
          block = [];
          break;

          default:
          if (utils.isArray(ast)) {
            str += this.getBlock(ast);
          } else {
            str += ast;
          }
          break;
        }
      }, this);

      return str;
    }
  });
};
